/*!@file
** @brief   The abstract representation of a file or a directory.
** @remarks More features for Windows Template Library are provided.
** @author  UETA, Shin-ichi <ueta@usdesign.jp>
**
** @copyright (C) 2014-2018 Ueta System Design Studio\n
** This file is licensed under the terms of the MIT License.\n
** See <https://opensource.org/licenses/MIT>.
*/
#pragma once

/*!
** @brief   The abstract representation of a file or a directory.
** @remarks -
*/
class CFile : public CPath
{
public:

	/*!
	** @brief   Initializes a new CFile object.
	** @remarks -
	*/
	CFile()
	{
		;
	}

	/*!
	** @brief   Constructs a new CFile object from the specified path.
	** @param   [in] path the path to a file or a directory to assign.
	** @remarks -
	*/
	CFile(const CPath& path) : CPath(path)
	{
		;
	}

	/*!
	** @brief   Assigns the path to a file or a directory.
	** @param   [in] path the path to a file or a directory to assign.
	** @returns By convention, a reference to this object.
	** @remarks -
	*/
	CFile& operator =(const CPath& path)
	{
		CPath::operator =(path);

		return *this;
	}

	/*!
	** @brief   Returns the abstract representation of the parent directory.
	** @returns A instance of the CFile class that represents the parent.
	** @remarks -
	*/
	CFile GetParent() const
	{
		CPath path = *this;
		path.RemoveFileSpec();
		return path;
	}

	/*!
	** @brief   Ensures that the file exists at the path.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL EnsureFileExists()
	{
		if (FileExists())
		{
			return TRUE;
		}
		HANDLE hFile = ::CreateFile(
				*this,
				GENERIC_WRITE,
				FILE_SHARE_READ | FILE_SHARE_WRITE,
				NULL,
				OPEN_ALWAYS,
				FILE_ATTRIBUTE_NORMAL,
				NULL);

		if (hFile != INVALID_HANDLE_VALUE)
		{
			::CloseHandle(hFile);

			return TRUE;
		}
		return FALSE;
	}

	/*!
	** @brief   Retrieves information about the file or directory.
	** @param   [out] sfi a SHFILEINFO structure to receive the file information.
	** @param   [in] uFlags specifies the file information to retrieve.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL GetFileInfo(SHFILEINFO& sfi, UINT uFlags)
	{
		if (::SHGetFileInfo(*this, FILE_ATTRIBUTE_NORMAL, &sfi, sizeof(SHFILEINFO), uFlags)
		 || ::SHGetFileInfo(*this, FILE_ATTRIBUTE_NORMAL, &sfi, sizeof(SHFILEINFO), uFlags | SHGFI_USEFILEATTRIBUTES))
		{
			return TRUE;
		}
		return FALSE;
	}

	/*!
	** @brief   Retrieves the file times for the file or directory.
	** @param   [out] lpCreationTime a pointer to a FILETIME structure to receive the creation time.
	** @param   [out] lpLastAccessTime a pointer to a FILETIME structure to receive the access time.
	** @param   [out] lpLastWriteTime a pointer to a FILETIME structure to receive the modification time.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL GetFileTime(FILETIME* lpCreationTime, FILETIME* lpLastAccessTime, FILETIME* lpLastWriteTime)
	{
		HANDLE hFile = ::CreateFile(
				*this,
				FILE_READ_ATTRIBUTES,
				FILE_SHARE_READ | FILE_SHARE_WRITE,
				NULL,
				OPEN_EXISTING,
				FILE_FLAG_BACKUP_SEMANTICS | FILE_FLAG_OPEN_REPARSE_POINT,
				NULL);

		if (hFile != INVALID_HANDLE_VALUE)
		{
			BOOL bSuccess = ::GetFileTime(hFile, lpCreationTime, lpLastAccessTime, lpLastWriteTime);

			::CloseHandle(hFile);

			return bSuccess;
		}
		return FALSE;
	}

	/*!
	** @brief   Sets the file times for the file or directory.
	** @param   [in] lpCreationTime a pointer to a FILETIME structure that contains the creation time.
	** @param   [in] lpLastAccessTime a pointer to a FILETIME structure that contains the access time.
	** @param   [in] lpLastWriteTime a pointer to a FILETIME structure that contains the modification time.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL SetFileTime(const FILETIME* lpCreationTime, const FILETIME* lpLastAccessTime, const FILETIME* lpLastWriteTime)
	{
		HANDLE hFile = ::CreateFile(
				*this,
				FILE_WRITE_ATTRIBUTES,
				FILE_SHARE_READ | FILE_SHARE_WRITE,
				NULL,
				OPEN_EXISTING,
				FILE_FLAG_BACKUP_SEMANTICS | FILE_FLAG_OPEN_REPARSE_POINT,
				NULL);

		if (hFile != INVALID_HANDLE_VALUE)
		{
			BOOL bSuccess = ::SetFileTime(hFile, lpCreationTime, lpLastAccessTime, lpLastWriteTime);

			::CloseHandle(hFile);

			return bSuccess;
		}
		return FALSE;
	}

	/*!
	** @brief   Checks to see if the path exists.
	** @param   [in] hWndOwner a handle to the owner window to be used for any user interface windows.
	** @param   [in] dwFlags determines behavior options.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL PrepareForWrite(HWND hWndOwner, DWORD dwFlags = SHPPFW_DEFAULT)
	{
		HRESULT hResult = ::SHPathPrepareForWrite(hWndOwner, NULL, *this, dwFlags);

		if (SUCCEEDED(hResult))
		{
			return TRUE;
		}
		return FALSE;
	}
};

/*!
** @brief   The date and time values associated with a file.
** @remarks -
*/
class CFileTimeEx : public CFileTime
{
public:

	/*!
	** @brief   Initializes a new CFileTimeEx object.
	** @remarks -
	*/
	CFileTimeEx() : CFileTime(MAXULONGLONG)
	{
		;
	}

	/*!
	** @brief   Constructs a new CFileTimeEx object from the specified date and time.
	** @param   [in] ft a FILETIME structure that contains the date and time to assign.
	** @remarks -
	*/
	CFileTimeEx(const FILETIME& ft) : CFileTime(ft)
	{
		;
	}

	/*!
	** @brief   Constructs a new CFileTimeEx object from the specified date and time.
	** @param   [in] nTime the date and time expressed as a 64-bit value to assign.
	** @remarks -
	*/
	CFileTimeEx(ULONGLONG nTime) : CFileTime(nTime)
	{
		;
	}

	/*!
	** @brief   Retrieves the current system date and time.
	** @returns A FILETIME structure that contains the UTC-based file time.
	** @remarks -
	*/
	static CFileTimeEx GetCurrentTime()
	{
		return CFileTime::GetCurrentTime();
	}

	/*!
	** @brief   Assigns the specified value.
	** @param   [in] ft a FILETIME structure that contains the date and time to assign.
	** @returns By convention, a reference to this object.
	** @remarks -
	*/
	CFileTimeEx& operator =(const FILETIME& ft)
	{
		return static_cast<CFileTimeEx&>(CFileTime::operator =(ft));
	}

	/*!
	** @brief   Adds a time-span to the date and time and assigns the result to this object.
	** @param   [in] span a time-span expressed as a 64-bit value to be added.
	** @returns By convention, a reference to this object.
	** @remarks -
	*/
	CFileTimeEx& operator +=(CFileTimeSpan span)
	{
		return operator =(operator +(span));
	}

	/*!
	** @brief   Subtracts a time-span from the date and time and assigns the result to this object.
	** @param   [in] span a time-span expressed as a 64-bit value to be subtracted.
	** @returns By convention, a reference to this object.
	** @remarks -
	*/
	CFileTimeEx& operator -=(CFileTimeSpan span)
	{
		return operator =(operator -(span));
	}

	/*!
	** @brief   Adds a time-span to the date and time.
	** @param   [in] span a time-span expressed as a 64-bit value to be added.
	** @returns A FILETIME structure that contains the result of the operation.
	** @remarks -
	*/
	CFileTimeEx operator +(CFileTimeSpan span) const
	{
		ULONGLONG t = GetTime();

		if (t == MAXULONGLONG)
			return CFileTimeEx(); // safeguard to preserve file time

		return CFileTimeEx(t + span.GetTimeSpan());
	}

	/*!
	** @brief   Subtracts a time-span from the date and time.
	** @param   [in] span a time-span expressed as a 64-bit value to be subtracted.
	** @returns A FILETIME structure that contains the result of the operation.
	** @remarks -
	*/
	CFileTimeEx operator -(CFileTimeSpan span) const
	{
		ULONGLONG t = GetTime();

		if (t == MAXULONGLONG)
			return CFileTimeEx(); // safeguard to preserve file time

		return CFileTimeEx(t - span.GetTimeSpan());
	}

	/*!
	** @brief   Converts the file time based on UTC to the local file time.
	** @returns A FILETIME structure that contains the local file time.
	** @remarks -
	*/
	CFileTimeEx UTCToLocal() const
	{
		CFileTimeEx ftLocalFileTime;

		#if 1 // daylight saving time aware
		{
			SYSTEMTIME stSystemTime;

			if (::FileTimeToSystemTime(this, &stSystemTime))
			{
				SYSTEMTIME stLocalTime;

				if (::SystemTimeToTzSpecificLocalTime(NULL, &stSystemTime, &stLocalTime))
				{
					if (::SystemTimeToFileTime(&stLocalTime, &ftLocalFileTime))
					{
						;
					}
				}
			}
		}
		#else

		::FileTimeToLocalFileTime(this, &ftLocalFileTime);

		#endif
		return ftLocalFileTime;
	}

	/*!
	** @brief   Converts the local file time to the file time based on UTC.
	** @returns A FILETIME structure that contains the UTC-based file time.
	** @remarks -
	*/
	CFileTimeEx LocalToUTC() const
	{
		CFileTimeEx ftFileTime;

		#if 1 // daylight saving time aware
		{
			SYSTEMTIME stLocalTime;

			if (::FileTimeToSystemTime(this, &stLocalTime))
			{
				SYSTEMTIME stSystemTime;

				if (::TzSpecificLocalTimeToSystemTime(NULL, &stLocalTime, &stSystemTime))
				{
					if (::SystemTimeToFileTime(&stSystemTime, &ftFileTime))
					{
						;
					}
				}
			}
		}
		#else

		::LocalFileTimeToFileTime(this, &ftFileTime);

		#endif
		return ftFileTime;
	}

	/*!
	** @brief   Retrieves the midnight of the day based on the time-zone.
	** @returns A FILETIME structure that contains the UTC-based file time.
	** @remarks -
	*/
	CFileTimeEx GetTzDate() const
	{
		ULONGLONG t = UTCToLocal().GetTime();

		if (t == MAXULONGLONG)
			return CFileTimeEx(); // safeguard to preserve file time

		return CFileTimeEx(t - t % CFileTime::Day).LocalToUTC();
	}

	/*!
	** @brief   Retrieves the time of the day based on the time-zone.
	** @returns A time-span expressed as a 64-bit value.
	** @remarks -
	*/
	CFileTimeSpan GetTzTime() const
	{
		ULONGLONG t = UTCToLocal().GetTime();

		ATLASSERT(t != MAXULONGLONG); // ought to be safe for both UTC-based time and local time

		return CFileTimeSpan(t % CFileTime::Day);
	}

	/*!
	** @brief   Retrieves the date and time value set the second to zero.
	** @returns A FILETIME structure that contains the UTC-based file time.
	** @remarks -
	*/
	CFileTimeEx ZeroSecond() const
	{
		ULONGLONG t = GetTime();

		ATLASSERT(t != MAXULONGLONG);

		return CFileTimeEx(t - t % CFileTime::Minute);
	}

	/*!
	** @brief   Retrieves the date and time value set the millisecond to zero.
	** @returns A FILETIME structure that contains the UTC-based file time.
	** @remarks -
	*/
	CFileTimeEx ZeroMillisecond() const
	{
		ULONGLONG t = GetTime();

		ATLASSERT(t != MAXULONGLONG);

		return CFileTimeEx(t - t % CFileTime::Second);
	}
};

/*!
** @brief   The feature that searches a directory for files or subdirectories.
** @remarks -
*/
class CFindFileEx : public CFindFile
{
public:

	/*!
	** @brief   Initializes a new CFindFileEx object.
	** @remarks -
	*/
	CFindFileEx()
	{
		;
	}

	/*!
	** @brief   Opens a file search.
	** @param   [in] pszDir the path to a directory to search.
	** @param   [in] pszFileSpec a null-terminated string that contains the file name pattern.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL FindFile(LPCTSTR pszDir, LPCTSTR pszFileSpec)
	{
		CPath path = pszDir;
		path += pszFileSpec;

		return CFindFile::FindFile(path);
	}

	/*!
	** @brief   Matches the name of the found file against one or more file name patterns.
	** @param   [in] pszSpec a null-terminated string that contains the file name pattern.
	** @param   [in] dwFlags modifies the search condition.
	** @returns TRUE if matches, or FALSE otherwise.
	** @remarks -
	*/
	BOOL MatchesSpec(LPCTSTR pszSpec, DWORD dwFlags = PMSF_NORMAL) const
	{
		if (m_bFound)
		{
			HRESULT hResult = ::PathMatchSpecEx(m_fd.cFileName, pszSpec, dwFlags);

			switch (hResult)
			{
			case S_OK:
				return TRUE;
			default:
				break;
			}
		}
		return FALSE;
	}
};
